import numpy as np
import xarray as xr


def _make_time_coord(
    time_start="2019-10-03T19:00:00.000",
    time_delta=3600,
    n_samples=10,
    time_scale="utc",
):
    from astropy.timeseries import TimeSeries
    from astropy.time import Time
    from astropy import units as u

    ts = np.array(
        TimeSeries(
            time_start=time_start,
            time_delta=time_delta * u.s,
            n_samples=n_samples,
        ).time.unix
    )

    return {
        "dims": "time",
        "data": ts,
        "attrs": {
            "units": "s",
            "type": "time",
            "format": "unix",
            "time_scale": time_scale,
        },
    }


def _make_frequency_coord(
    freq_start=3 * 10**9,
    freq_delta=0.4 * 10**9,
    n_channels=50,
    velocity_frame="lsrk",
):
    freq_chan = (np.arange(0, n_channels) * freq_delta + freq_start).astype(float)
    return {
        "dims": "frequency",
        "data": freq_chan,
        "attrs": {
            "units": "Hz",
            "type": "spectral_coord",
            "velocity_frame": velocity_frame,
        },
    }


def array_split(data, n_chunks):
    chunk_size = int(np.ceil(len(data) / n_chunks))
    
    from itertools import islice
    data_iter = iter(data)
    
    result = []
    for _ in range(n_chunks):
        chunk = list(islice(data_iter, chunk_size))
        if not chunk:
            break
        result.append(np.array(chunk))
    
    return result
 

def _make_parallel_coord(coord, n_chunks=None):
    if isinstance(coord, xr.core.dataarray.DataArray):
        coord = coord.copy(deep=True).to_dict()

    n_samples = len(coord["data"])
    parallel_coord = {}
    parallel_coord["data"] = coord["data"]
#    parallel_coord["data_chunks"] = dict(
#        zip(np.arange(n_chunks), np.array_split(coord["data"], n_chunks))
#    )
    parallel_coord["data_chunks"] = dict(
        zip(np.arange(n_chunks), array_split(coord["data"], n_chunks))
    )
    parallel_coord["dims"] = coord["dims"]
    parallel_coord["attrs"] = coord["attrs"]
    return parallel_coord



from xradio.vis.read_processing_set import read_processing_set
from xradio.image import read_image
import itertools, functools, operator
import numpy as np
import dask
import math
import os
import datetime
from astroviper._utils._logger import _get_logger
import ipaddress


def get_unique_resource_ip(workers_info):
    nodes = []
    for worker, wi in workers_info.items():
        worker_ip = worker[worker.rfind("/") + 1 : worker.rfind(":")]
        assert worker_ip in list(
            wi["resources"].keys()
        ), "local_cache enabled but workers have not been annotated. Make sure that local_cache has been set to True during client setup."
        if worker_ip not in nodes:
            nodes.append(worker_ip)
    return nodes


def _make_iter_chunks_indxs(parallel_coords):
    parallel_dims = []
    list_chunk_indxs = []
    n_chunks = 1
    for dim, pc in parallel_coords.items():
        chunk_indxs = list(pc["data_chunks"].keys())
        n_chunks = n_chunks * len(chunk_indxs)
        list_chunk_indxs.append(chunk_indxs)
        parallel_dims.append(dim)

    iter_chunks_indxs = itertools.product(*list_chunk_indxs)
    return iter_chunks_indxs, parallel_dims


def _generate_chunk_slices(parallel_coords, input_data_name):
    """ """
    from scipy.interpolate import interp1d

    # Construct an interpolator for each parallel dim:
    interp1d_dict = {}
    for dim, pc in parallel_coords.items():
        interp1d_dict[dim] = interp1d(
            pc["data"],
            np.arange(len(pc["data"])),
            kind="nearest",
            bounds_error=False,
            fill_value=-1
            # fill_value="extrapolate",
            # assume_sorted=True
        )  # Should we use fill_value='extrapolate' in interp1d?

    chunk_slice_dict = {}
    for xds_key in input_data_name:
        for dim, pc in parallel_coords.items():
            #print(input_data_name[xds_key],dim)
            interp_indx = interp1d_dict[dim](input_data_name[xds_key][dim].values).astype(int)
            # print(dim,interp_indx,len(interp_indx))

            # print(pc['data'][interp_indx])

            pc_chunk_start = 0
            pc_chunk_end = 0
            chunk_indx_start_stop = {}
            for chunk_key in np.arange(
                len(pc["data_chunks"])
            ):  # ensure that keys are ordered.
                pc_chunk_start = pc_chunk_end
                pc_chunk_end = len(pc["data_chunks"][chunk_key]) + pc_chunk_end

                start_slice = np.where(interp_indx == pc_chunk_start)[0]
                if start_slice.size != 0:
                    start_slice = start_slice[0]
                else:
                    start_slice = None

                end_slice = np.where(interp_indx == pc_chunk_end - 1)[0]

                if end_slice.size != 0:
                    end_slice = end_slice[-1] + 1
                else:
                    end_slice = len(interp_indx)

                if start_slice is None:
                    chunk_indx_start_stop[chunk_key] = slice(None)
                else:
                    if start_slice > end_slice:
                        chunk_indx_start_stop[chunk_key] = slice(end_slice, start_slice)
                    else:
                        chunk_indx_start_stop[chunk_key] = slice(start_slice, end_slice)

            if xds_key in chunk_slice_dict:
                chunk_slice_dict[xds_key][dim] = chunk_indx_start_stop
            else:
                chunk_slice_dict[xds_key] = {dim: chunk_indx_start_stop}

    return chunk_slice_dict


def _map(input_data_name, input_data_type, parallel_coords, func_chunk, input_parms, ps_sel_parms={}, client=None):
    """
    Builds a perfectly parallel graph where func_chunk node task is created for each chunk defined in parallel_coords. The data in the input_data_name is mapped to each parallel_coords chunk.
    """

    logger = _get_logger()
    
    if input_data_type=="processing_set":
        input_data = read_processing_set(input_data_name, ps_sel_parms["intents"], ps_sel_parms["fields"])
    elif input_data_type=="image":
        input_data = {'image':read_image(input_data_name)}

    iter_chunks_indxs, parallel_dims = _make_iter_chunks_indxs(parallel_coords)

    chunk_slice_dict = _generate_chunk_slices(parallel_coords, input_data)
    # print(chunk_slice_dict)

    #input_parms = {"input_data_name": input_data_name}
    input_parms["input_data_name"] = input_data_name

    if "VIPER_LOCAL_DIR" in os.environ:
        local_cache = True
        input_parms["viper_local_dir"] = os.environ["VIPER_LOCAL_DIR"]

        if "date_time" in ps_sel_parms:
            input_parms["date_time"] = ps_sel_parms["date_time"]
        else:
            input_parms["date_time"] = datetime.datetime.now().strftime("%y%m%d%H%M%S")
    else:
        local_cache = False
        input_parms["viper_local_dir"] = None
        input_parms["date_time"] = None

    if local_cache:
        workers_info = client.scheduler_info()["workers"]
        nodes_ip_list = get_unique_resource_ip(workers_info)
        n_nodes = len(nodes_ip_list)

        chunks_per_node = math.floor(n_chunks / n_nodes + 0.5)
        if chunks_per_node == 0:
            chunks_per_node = 1
        chunk_to_node_map = np.repeat(np.arange(n_nodes), chunks_per_node)

        if len(chunk_to_node_map) < n_chunks:
            n_pad = n_chunks - len(chunk_to_node_map)
            chunk_to_node_map = np.concatenate(
                [chunk_to_node_map, np.array([chunk_to_node_map[-1]] * n_pad)]
            )

    graph_list = []
    for i_chunk, chunk_indx in enumerate(iter_chunks_indxs):
        # print("chunk_indx", i_chunk, chunk_indx)

        single_chunk_slice_dict = {}
        for xds_id in input_data.keys():
            single_chunk_slice_dict[xds_id] = {}
            empty_chunk = False
            for i, chunk_id in enumerate(chunk_indx):
                if chunk_id in chunk_slice_dict[xds_id][parallel_dims[i]]:
                    single_chunk_slice_dict[xds_id][
                        parallel_dims[i]
                    ] = chunk_slice_dict[xds_id][parallel_dims[i]][chunk_id]

                    if chunk_slice_dict[xds_id][parallel_dims[i]][chunk_id] == slice(
                        None
                    ):
                        empty_chunk = True
                else:
                    empty_chunk = True

            if (
                empty_chunk
            ):  # The xds with xds_id has no data for the parallel chunk (no slice on one of the dims).
                single_chunk_slice_dict.pop(xds_id, None)
        # print(single_chunk_slice_dict)
        if single_chunk_slice_dict:
            input_parms["data_sel"] = single_chunk_slice_dict
            input_parms["chunk_coords"] = {}
            for i_dim, dim in enumerate(parallel_dims):
                chunk_coords = {}
                chunk_coords["data"] = parallel_coords[dim]["data_chunks"][
                    chunk_indx[i_dim]
                ]
                chunk_coords["dims"] = parallel_coords[dim]["dims"]
                chunk_coords["attrs"] = parallel_coords[dim]["attrs"]
                input_parms["chunk_coords"][dim] = chunk_coords
            input_parms["chunk_indices"] = chunk_indx
            input_parms["chunk_id"] = chunk_id
            input_parms["parallel_dims"] = parallel_dims

            if local_cache:
                node_ip = nodes_ip_list[chunk_to_node_map[i_chunk]]
                logger.debug(
                    "Task with chunk id "
                    + str(chunk_id)
                    + " is assigned to ip "
                    + str(node_ip)
                )
                input_parms["node_ip"] = node_ip
                with dask.annotate(resources={node_ip: 1}):
                    graph_list.append(
                        dask.delayed(func_chunk)(dask.delayed(input_parms))
                    )
            else:
                # print("input_parms",input_parms)
                graph_list.append(dask.delayed(func_chunk)(dask.delayed(input_parms)))

        """
        a = np.ravel_multi_index([127,0,0,1],[256,256,256,256])
        #a = np.ravel_multi_index([255,255,255,254],[256,256,256,256])
        print(a)
        """

    return graph_list, input_parms["date_time"]



import xarray as xr


def _moment_max(img_xds_1, img_xds_2, input_parms):
    img_xds = xr.concat(
        [
            img_xds_1.max(dim="frequency", keepdims=True),
            img_xds_2.max(dim="frequency", keepdims=True),
        ],
        dim="frequency",
    )
    return img_xds.max(dim="frequency", keepdims=True)



def make_empty_sky_image(
    xds,
    phase_center,
    image_size,
    cell_size,
    chan_coords,
    chan_width,
    pol_coords,
    time_coords,
    direction_reference="FK5",
    projection="SIN",
    spectral_reference="lsrk",
    velocity_type="radio",
    unit="Jy/beam",
    data_group_name="base",
):
    """
    Create an img_dataset with only coordinates (no datavariables).
    The image dimensionality is either:
        l, m, time, chan, pol

    Parameters
    ----------
    xds : xarray.Dataset
        Empty dataset (dataset = xarray.Dataset())
    phase_center : array of number, length = 2, units = rad
        Image phase center.
    image_size : array of int, length = 2, units = rad
        Number of x and y axis pixels in image.
    cell_size : array of number, length = 2, units = rad
        Cell size of x and y axis pixels in image.
    chan_coords : dask.Array
        The center frequency of each image channel.
    chan_width : dask.Array
        The frequency width of each image channel.
    pol_coords : dask.Array
        The polarization code for each image polarization.
    time_coords : dask.Array
        The time for each image time step.
    direction_reference : str, default = 'FK5'
    projection : str, default = 'SIN'
    spectral_reference : str, default = 'lsrk'
    velocity_type : str, default = 'radio'
    Returns
    -------
    xarray.Dataset
        new xarray Datasets
    """
    import xarray as xr
    import numpy as np

    from astropy.wcs import WCS

    rad_to_deg = 180 / np.pi
    w = WCS(naxis=2)
    w.wcs.crpix = np.array(image_size) // 2
    w.wcs.cdelt = np.array(cell_size) * rad_to_deg
    w.wcs.crval = np.array(phase_center) * rad_to_deg
    w.wcs.ctype = ["RA---" + projection, "DEC--" + projection]

    x = np.arange(image_size[0])
    y = np.arange(image_size[1])
    X, Y = np.meshgrid(x, y, indexing="ij")
    ra, dec = w.wcs_pix2world(X, Y, 1)

    # The l,m value can be caluclated from non-integer index using l_val = cell_size*l_indx - image_center[0]*cell_size[0]. This arises when using wcs libarary all_world2pix
    image_center = np.array(image_size) // 2
    l = np.arange(-image_center[0], image_size[0] - image_center[0]) * cell_size[0]
    m = np.arange(-image_center[1], image_size[1] - image_center[1]) * cell_size[1]

    coords = {
        "time": time_coords,
        "frequency": chan_coords,
        "pol": pol_coords,
        "l": l,
        "m": m,
        "right_ascension": (("l", "m"), ra / rad_to_deg),
        "declination": (("l", "m"), dec / rad_to_deg),
    }
    # coords = {'time':time_coords,'chan': chan_coords, 'pol': pol_coords, 'chan_width' : ('chan',chan_width),'l':l,'m':m,'right_ascension' : (('l','m'),ra/rad_to_deg),'declination' : (('l','m'),dec/rad_to_deg)}
    xds.attrs["axis_units"] = ["rad", "rad", "time", "Hz", "pol"]

    xds = xds.assign_coords(coords)

    xds.attrs["direction_reference"] = direction_reference
    xds.attrs["spectral_reference"] = spectral_reference
    xds.attrs["velocity_type"] = velocity_type
    xds.attrs["unit"] = unit

    xds.attrs["data_groups"] = {data_group_name: {}}

    return xds



import warnings, time, os, psutil, multiprocessing, re
import dask
import copy
import os
import logging
import astroviper
import distributed
from astroviper._utils._parm_utils._check_logger_parms import (
    _check_logger_parms,
    _check_worker_logger_parms,
)
from astroviper._utils._logger import (
    _setup_logger,
    _get_logger,
    _setup_logger,
)
from astroviper._concurrency._dask._worker import (
    _worker,
)  # _worker_logger_plugin


def local_client(
    cores=None,
    memory_limit=None,
    autorestrictor=False,
    dask_local_dir=None,
    local_dir=None,
    wait_for_workers=True,
    log_parms={},
    worker_log_parms={},
):
    """
    local_dir setting is only useful for testing since this function creates a local cluster. slurm_cluster_client should be used for a multinode cluster.

    https://github.com/dask/dask/issues/5577
    log_parms['log_to_term'] = True/False
    log_parms['log_file'] = True/False
    log_parms['log_level'] =
    """

    _log_parms = copy.deepcopy(log_parms)
    _worker_log_parms = copy.deepcopy(worker_log_parms)

    assert _check_logger_parms(
        _log_parms
    ), "######### ERROR: initialize_processing log_parms checking failed."

    if _worker_log_parms is not None:
        assert _check_worker_logger_parms(
            _worker_log_parms
        ), "######### ERROR: initialize_processing log_parms checking failed."

    if local_dir:
        os.environ["VIPER_LOCAL_DIR"] = local_dir
        local_cache = True
    else:
        local_cache = False

    print(_log_parms)
    _setup_logger(**_log_parms)
    logger = _get_logger()

    _set_up_dask(dask_local_dir)

    # viper_path = astroviper.__path__.__dict__["_path"][0]
    viper_path = astroviper.__path__[0]
    if local_cache or autorestrictor:
        dask.config.set(
            {
                "distributed.scheduler.preload": os.path.join(
                    viper_path, "_concurrency/_dask/_scheduler.py"
                )
            }
        )
        dask.config.set(
            {
                "distributed.scheduler.preload-argv": [
                    "--local_cache",
                    local_cache,
                    "--autorestrictor",
                    autorestrictor,
                ]
            }
        )

    """ This method of assigning a worker plugin does not seem to work when using dask_jobqueue. Consequently using client.register_worker_plugin so that the method of assigning a worker plugin is the same for local_client and slurm_cluster_client.
    if local_cache or _worker_log_parms:
        dask.config.set({"distributed.worker.preload": os.path.join(viper_path,'_utils/_worker.py')})
        dask.config.set({"distributed.worker.preload-argv": ["--local_cache",local_cache,"--log_to_term",_worker_log_parms['log_to_term'],"--log_to_file",_worker_log_parms['log_to_file'],"--log_file",_worker_log_parms['log_file'],"--log_level",_worker_log_parms['log_level']]})
    """
    # setup distributed based multiprocessing environment
    if cores is None:
        cores = multiprocessing.cpu_count()
    if memory_limit is None:
        memory_limit = (
            str(round(((psutil.virtual_memory().available / (1024**2))) / cores))
            + "MB"
        )
    cluster = distributed.LocalCluster(
        n_workers=cores, threads_per_worker=1, processes=True, memory_limit=memory_limit
    )  # , silence_logs=logging.ERROR #,resources={'GPU': 2}
    client = distributed.Client(cluster)
    client.get_versions(check=True)

    """
    When constructing a graph that has local cache enabled all workers need to be up and running.
    """
    if local_cache or wait_for_workers:
        client.wait_for_workers(n_workers=cores)

    if local_cache or _worker_log_parms:
        plugin = _worker(local_cache, _worker_log_parms)
        client.register_worker_plugin(plugin, name="viper_worker")

    logger.info("Created client " + str(client))

    return client


def slurm_cluster_client(
    workers_per_node,
    cores_per_node,
    memory_per_node,
    number_of_nodes,
    queue,
    interface,
    python_env_dir,
    dask_local_dir,
    dask_log_dir,
    exclude_nodes="nmpost090",
    dashboard_port=9000,
    local_dir=None,
    autorestrictor=False,
    wait_for_workers=True,
    log_parms={},
    worker_log_parms={},
):
    """
    local_cache setting is only useful for testing since this function creates a local cluster. slurm_cluster_client should be used for a multinode cluster.

    https://github.com/dask/dask/issues/5577
    log_parms['log_to_term'] = True/False
    log_parms['log_file'] = True/False
    log_parms['log_level'] =

    interface eth0, ib0
    python "/mnt/condor/jsteeb/viper_py/bin/python"
    dask_local_dir "/mnt/condor/jsteeb"
    dask_log_dir "/.lustre/aoc/projects/ngvla/viper/ngvla_sim",
    """

    from dask_jobqueue import SLURMCluster
    from distributed import Client, config, performance_report

    _log_parms = copy.deepcopy(log_parms)
    _worker_log_parms = copy.deepcopy(worker_log_parms)

    assert _check_logger_parms(
        _log_parms
    ), "######### ERROR: initialize_processing log_parms checking failed."
    assert _check_worker_logger_parms(
        _worker_log_parms
    ), "######### ERROR: initialize_processing log_parms checking failed."

    if local_dir:
        os.environ["VIPER_LOCAL_DIR"] = local_dir
        local_cache = True
    else:
        local_cache = False

    # Viper logger for code that is not part of the Dask graph. The worker logger is setup in the _viper_worker plugin.
    from viper._utils._logger import _setup_logger

    _setup_logger(**_log_parms)
    logger = _get_logger()

    _set_up_dask(dask_local_dir)

    viper_path = astroviper.__path__.__dict__["_path"][0]
    if local_cache or autorestrictor:
        dask.config.set(
            {
                "distributed.scheduler.preload": os.path.join(
                    viper_path, "_concurrency/_dask/_scheduler.py"
                )
            }
        )
        dask.config.set(
            {
                "distributed.scheduler.preload-argv": [
                    "--local_cache",
                    local_cache,
                    "--autorestrictor",
                    autorestrictor,
                ]
            }
        )

    """ This method of assigning a worker plugin does not seem to work when using dask_jobqueue. Consequently using client.register_worker_plugin so that the method of assigning a worker plugin is the same for local_client and slurm_cluster_client.
    if local_cache or _worker_log_parms:
        dask.config.set({"distributed.worker.preload": os.path.join(viper_path,'_utils/_worker.py')})
        dask.config.set({"distributed.worker.preload-argv": ["--local_cache",local_cache,"--log_to_term",_worker_log_parms['log_to_term'],"--log_to_file",_worker_log_parms['log_to_file'],"--log_file",_worker_log_parms['log_file'],"--log_level",_worker_log_parms['log_level']]})
    """

    cluster = SLURMCluster(
        processes=workers_per_node,
        cores=cores_per_node,
        interface=interface,
        memory=memory_per_node,
        walltime="24:00:00",
        queue=queue,
        name="viper",
        python=python_env_dir,  # "/mnt/condor/jsteeb/viper_py/bin/python", #"/.lustre/aoc/projects/ngvla/viper/viper_py_env/bin/python",
        local_directory=dask_local_dir,  # "/mnt/condor/jsteeb",
        log_directory=dask_log_dir,
        job_extra_directives=["--exclude=" + exclude_nodes],
        # job_extra_directives=["--exclude=nmpost087,nmpost089,nmpost088"],
        scheduler_options={"dashboard_address": ":" + str(dashboard_port)},
    )  # interface='ib0'

    client = Client(cluster)

    cluster.scale(workers_per_node * number_of_nodes)

    """
    When constructing a graph that has local cache enabled all workers need to be up and running.
    """
    if local_cache or wait_for_workers:
        client.wait_for_workers(n_workers=workers_per_node * number_of_nodes)

    if local_cache or _worker_log_parms:
        plugin = _worker(local_cache, _worker_log_parms)
        client.register_worker_plugin(plugin, name="viper_worker")

    logger.info("Created client " + str(client))

    return client


def _set_up_dask(local_directory):
    if local_directory:
        dask.config.set({"temporary_directory": local_directory})
    dask.config.set({"distributed.scheduler.allowed-failures": 10})
    dask.config.set({"distributed.scheduler.work-stealing": True})
    dask.config.set({"distributed.scheduler.unknown-task-duration": "99m"})
    dask.config.set({"distributed.worker.memory.pause": False})
    dask.config.set({"distributed.worker.memory.terminate": False})
    # dask.config.set({"distributed.worker.memory.recent-to-old-time": '999s'})
    dask.config.set({"distributed.comm.timeouts.connect": "3600s"})
    dask.config.set({"distributed.comm.timeouts.tcp": "3600s"})
    dask.config.set({"distributed.nanny.environ.OMP_NUM_THREADS": 1})
    dask.config.set({"distributed.nanny.environ.MKL_NUM_THREADS": 1})
    # https://docs.dask.org/en/stable/how-to/customize-initialization.html



import click

from astroviper._utils._logger import _setup_worker_logger


class _worker:
    def __init__(self, local_cache, log_parms):
        print("init local cache")
        self.local_cache = local_cache

        print("log_parms", log_parms)
        # /.lustre/aoc/projects/ngvla/viper/ngvla_sim/viper_
        self.log_to_term = log_parms["log_to_term"]
        self.log_to_file = log_parms["log_to_file"]
        self.log_file = log_parms["log_file"]
        self.log_level = log_parms["log_level"]

    def get_logger(self):
        return self.logger

    def setup(self, worker):
        """
        Run when the plugin is attached to a worker. This happens when the plugin is registered
        and attached to existing workers, or when a worker is created after the plugin has been
        registered.
        """

        self.logger = _setup_worker_logger(
            self.log_to_term, self.log_to_file, self.log_file, self.log_level
        )
        self.logger.debug(
            "Logger created on worker " + str(worker.id) + ",*," + str(worker.address)
        )

        # Documentation https://distributed.dask.org/en/stable/worker.html#distributed.worker.Worker
        self.worker = worker

        if self.local_cache:
            ip = worker.address[
                worker.address.rfind("/") + 1 : worker.address.rfind(":")
            ]
            self.logger.debug(str(worker.id) + ",*," + ip)
            worker.state.available_resources = {
                **worker.state.available_resources,
                **{ip: 1},
            }
            # print(worker.state.available_resources)


# https://github.com/dask/distributed/issues/4169
@click.command()
@click.option("--local_cache", default=False)
# @click.option("--log_parms", default={'log_to_term':True,'log_to_file':False,'log_file':'viper_', 'log_level':'DEBUG'})
@click.option("--log_to_term", default=True)
@click.option("--log_to_file", default=False)
@click.option("--log_file", default="viper_")
@click.option("--log_level", default="INFO")
async def dask_setup(
    worker, local_cache, log_to_term, log_to_file, log_file, log_level
):
    log_parms = {
        "log_to_term": log_to_term,
        "log_to_file": log_to_file,
        "log_file": log_file,
        "log_level": log_level,
    }
    plugin = _worker(local_cache, log_parms)
    await worker.client.register_worker_plugin(plugin, name="viper_worker")



logger_name = "viper"
import sys
import logging
from datetime import datetime

formatter = logging.Formatter(
    "[%(filename)s:%(lineno)s - %(funcName)20s() ] %(message)s"
)
from dask.distributed import get_worker
from dask.distributed import WorkerPlugin
import dask


def _get_logger(name=logger_name):
    """
    Will first try to get worker logger. If this fails graph construction logger is returned.
    """
    from dask.distributed import get_worker

    try:
        worker = get_worker()
    except:
        return logging.getLogger(name)

    try:
        logger = worker.plugins["viper_worker"].get_logger()
        return logger
    except:
        return logging.getLogger()


def _setup_logger(
    log_to_term=False,
    log_to_file=True,
    log_file="viper_",
    log_level="INFO",
    name=logger_name,
):
    """To setup as many loggers as you want"""
    logger = logging.getLogger(name)
    logger.setLevel(logging.getLevelName(log_level))

    if log_to_term:
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    if log_to_file:
        log_file = log_file + datetime.today().strftime("%Y%m%d_%H%M%S") + ".log"
        handler = logging.FileHandler(log_file)
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    return logger


def _get_worker_logger_name(name=logger_name):
    worker_log_name = name + "_" + str(get_worker().id)
    return worker_log_name


"""
class _viper_worker_logger_plugin(WorkerPlugin):
    def __init__(self,log_parms):
        self.log_to_term=log_parms['log_to_term']
        self.log_to_file=log_parms['log_to_file']
        self.log_file=log_parms['log_file']
        self.level=log_parms['log_level']
        self.logger = None
        print(self.log_to_term,self.log_to_file,self.log_file,self.log_level)
        
    def get_logger(self):
        return self.logger
        
    def setup(self, worker: dask.distributed.Worker):
        "Run when the plugin is attached to a worker. This happens when the plugin is registered and attached to existing workers, or when a worker is created after the plugin has been registered."
        self.logger = _setup_viper_worker_logger(self.log_to_term,self.log_to_file,self.log_file,self.level)
"""


def _setup_worker_logger(log_to_term, log_to_file, log_file, log_level):
    parallel_logger_name = _get_worker_logger_name()

    logger = logging.getLogger(parallel_logger_name)
    logger.setLevel(logging.getLevelName(log_level))

    if log_to_term:
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    if log_to_file:
        log_file = (
            log_file
            + "_"
            + str(get_worker().id)
            + "_"
            + datetime.today().strftime("%Y%m%d_%H%M%S")
            + ".log"
        )
        handler = logging.FileHandler(log_file)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    return logger



def _tree_combine(list_to_combine, chunk_function, chunk_function_input_parms):
    import dask

    while len(list_to_combine) > 1:
        new_list_to_combine = []
        for i in range(0, len(list_to_combine), 2):
            if i < len(list_to_combine) - 1:
                lazy = dask.delayed(chunk_function)(
                    list_to_combine[i],
                    list_to_combine[i + 1],
                    chunk_function_input_parms,
                )
            else:
                lazy = list_to_combine[i]
            new_list_to_combine.append(lazy)
        list_to_combine = new_list_to_combine
    return list_to_combine


def _reduce(graph, chunk_function, chunk_function_input_parms, mode="tree"):
    if mode == "tree":
        graph_reduced = _tree_combine(
            graph[0], chunk_function, chunk_function_input_parms
        )

    return graph_reduced



def _wrap_func_with_dio(
    mds, chunk_function, chunk_function_input_parms, mds_sel_parms_chunk
):
    if mds is Dataset:
        mds_chunk = mds
    else:
        # Load the data from cache or common storage. If data loads from common storage it can be cache.
        mds_chunk = transfer_mds_chunk(mds_name, mds_sel_parms_chunk)

    # Run the chunk function on the above selected mds_chunk.
    mds_result_chunk = chunk_function(mds_chunk, chunk_function_input_parms)

    if chunk_function_input_parms["return_data"]:
        return mds_result_chunk
    else:
        ## Save the results
        transfer_mds_chunk(mds_result_chunk)
        results_dict = mds_result_chunk["results_dict"]
        return results_dict



from collections import defaultdict
from distributed import SchedulerPlugin
from dask.core import reverse_dict
from dask.base import tokenize
from dask.order import graph_metrics, ndependencies
import click
from distributed.diagnostics.plugin import SchedulerPlugin
import numpy as np


def unravel_deps(hlg_deps, name, unravelled_deps=None):
    """Recursively construct a set of all dependencies for a specific task."""

    if unravelled_deps is None:
        unravelled_deps = set()

    for dep in hlg_deps[name]:
        unravelled_deps |= {dep}
        unravel_deps(hlg_deps, dep, unravelled_deps)

    return unravelled_deps


def get_node_depths(dependencies, root_nodes, metrics):
    node_depths = {}

    for k in dependencies.keys():
        # Get dependencies per node.
        deps = unravel_deps(dependencies, k)
        # Associate nodes with root nodes.
        roots = root_nodes & deps
        offset = metrics[k][-1]
        node_depths[k] = max(metrics[r][-1] - offset for r in roots) if roots else 0

    return node_depths


class schedular(SchedulerPlugin):
    def __init__(self, autorestrictor, local_cache):
        self.autorestrictor = autorestrictor
        self.local_cache = local_cache
        super().__init__()

    def add_worker(self, scheduler, worker):
        if self.local_cache:
            # Set the resource label to the ip of the node that the worker is on, so that tasks that require a specific node can be assigned to the correct worker.
            ip = worker[worker.rfind("/") + 1 : worker.rfind(":")]
            scheduler.add_resources(worker=worker, resources={ip: 1})

    def update_graph(self, scheduler, dsk=None, keys=None, restrictions=None, **kw):
        if self.autorestrictor:
            print("Using autorestrictor")
            """Processes dependencies to assign tasks to specific workers."""
            workers = list(scheduler.workers.keys())
            n_worker = len(workers)

            tasks = scheduler.tasks
            dependencies = kw["dependencies"]

            # print('In update_graph :', scheduler, ',*,', dsk, ',*,', keys , ',*,', restrictions , ',*,', kw)
            if dependencies:
                dependents = reverse_dict(dependencies)

                # print('reversed dict:', dependents)

                _, total_dependencies = ndependencies(dependencies, dependents)
                # TODO: Avoid calling graph metrics.
                metrics = graph_metrics(dependencies, dependents, total_dependencies)

                # Terminal nodes have no dependents, root nodes have no dependencies.
                # Horizontal partition nodes are initialized as the terminal nodes.
                part_nodes = {k for (k, v) in dependents.items() if not v}
                root_nodes = {k for (k, v) in dependencies.items() if not v}

                # Figure out the depth of every task. Depth is defined as maximum
                # distance from a root node. TODO: Optimize get_node_depths.

                node_depths = get_node_depths(dependencies, root_nodes, metrics)
                # try:
                max_depth = max(node_depths.values())
                # except:
                #        print('&&&&& dependencies, root_nodes, metrics',node_depths,',*,',dependencies, root_nodes, metrics)

                # If we have fewer partition nodes than workers, we cannot utilise all
                # the workers and are likely dealing with a reduction. We work our way
                # back through the graph, starting at the deepest terminal nodes, and
                # try to find a depth at which there was enough work to utilise all
                # workers.
                while (len(part_nodes) < n_worker) & (max_depth > 0):
                    _part_nodes = part_nodes.copy()
                    for pn in _part_nodes:
                        if node_depths[pn] == max_depth:
                            part_nodes ^= set((pn,))
                            part_nodes |= dependencies[pn]
                    max_depth -= 1
                    if max_depth <= 0:
                        return  # In this case, there in nothing we can do - fall back.

                part_roots = {}
                part_dependencies = {}
                part_dependents = {}

                for pn in part_nodes:
                    # Get dependencies per partition node.
                    part_dependencies[pn] = unravel_deps(dependencies, pn)
                    # Get dependents per partition node.
                    part_dependents[pn] = unravel_deps(dependents, pn)
                    # Associate partition nodes with root nodes.
                    part_roots[pn] = root_nodes & part_dependencies[pn]

                # Create a unique token for each set of partition roots. TODO: This is
                # very strict. What about nodes with very similar roots? Tokenization
                # may be overkill too.
                root_tokens = {tokenize(*sorted(v)): v for v in part_roots.values()}

                hash_map = defaultdict(set)
                group_offset = 0

                # Associate partition roots with a specific group if they are not a
                # subset of another, larger root set.
                for k, v in root_tokens.items():
                    if any(v < vv for vv in root_tokens.values()):  # Strict subset.
                        continue
                    else:
                        hash_map[k] |= set([group_offset])
                        group_offset += 1

                # If roots were a subset, they should share the group of their
                # superset/s.
                for k, v in root_tokens.items():
                    if not v:  # Special case - no dependencies. Handled below.
                        continue
                    shared_roots = {
                        kk: None for kk, vv in root_tokens.items() if v < vv
                    }
                    if shared_roots:
                        hash_map[k] = set().union(
                            *[hash_map[kk] for kk in shared_roots.keys()]
                        )

                task_groups = defaultdict(set)

                for pn in part_nodes:
                    pdp = part_dependencies[pn]
                    pdn = part_dependents[pn]

                    if pdp:
                        groups = hash_map[tokenize(*sorted(part_roots[pn]))]
                    else:  # Special case - no dependencies.
                        groups = {group_offset}
                        group_offset += 1

                    for g in groups:
                        task_groups[g] |= pdp | pdn | {pn}

                worker_loads = {wkr: 0 for wkr in workers}

                for task_group in task_groups.values():
                    assignee = min(worker_loads, key=worker_loads.get)
                    worker_loads[assignee] += len(task_group)

                    for task_name in task_group:
                        try:
                            task = tasks[task_name]
                        except KeyError:  # Keys may not have an assosciated task.
                            continue

                        # print('^^^^^^',dir(task))
                        # if task._worker_restrictions is None:
                        #    task._worker_restrictions = set()
                        # task._worker_restrictions |= {assignee}
                        # task._loose_restrictions = False

                        if task.worker_restrictions is None:
                            task.worker_restrictions = set()
                        task.worker_restrictions |= {assignee}
                        task.loose_restrictions = False


@click.command()
@click.option("--autorestrictor", default=False)
@click.option("--local_cache", default=False)
def dask_setup(scheduler, autorestrictor, local_cache):
    plugin = schedular(autorestrictor, local_cache)
    scheduler.add_plugin(plugin)



def _make_image(input_parms):
    from xradio.vis.load_processing_set import load_processing_set

    ps = load_processing_set(
        ps_name=input_parms["input_data_name"], sel_parms=input_parms["data_sel"]
    )

    # print(ps.keys())

    import numpy as np
    from astroviper._domain._visibility._phase_shift import _phase_shift_vis_ds
    from astroviper._domain._imaging._make_imaging_weights import _make_imaging_weights
    from astroviper._domain._imaging._make_gridding_convolution_function import (
        _make_gridding_convolution_function,
    )
    from astroviper._domain._imaging._make_aperture_grid import _make_aperture_grid
    from astroviper._domain._imaging._make_uv_sampling_grid import (
        _make_uv_sampling_grid,
    )
    from astroviper.image.make_empty_sky_image import make_empty_sky_image
    from astroviper._domain._imaging._make_visibility_grid import _make_visibility_grid
    from astroviper._domain._imaging._fft_norm_img_xds import _fft_norm_img_xds

    from xradio.vis.load_processing_set import load_processing_set
    import xarray as xr

    # print(shift_parms)

    grid_parms = input_parms["grid_parms"]
    
    shift_parms = {}
#    shift_parms["new_phase_direction"] = ps[
#        "Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_12"
#    ].attrs["field_info"]["phase_direction"]
    shift_parms["new_phase_direction"] = grid_parms["phase_direction"]
    shift_parms["common_tangent_reprojection"] = True
#    grid_parms = {}
#    grid_parms["chan_mode"] = "cube"
#    grid_parms["image_size"] = [500, 500]
#    grid_parms["cell_size"] = np.array([-0.13, 0.13]) * np.pi / (180 * 3600)
#    grid_parms["fft_padding"] = 1.0
#    grid_parms["phase_direction"] = shift_parms["new_phase_direction"]
    # print('grid_parms[phase_direction]',grid_parms['phase_direction'])
    # print(ps['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_1'].attrs['field_info']['phase_dir'])

    # print(ps['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_1'])
    # print(ps['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_1'].frequency)

    #NB NB this needs to change
    freq_coords = ps[list(ps.keys())[0]].frequency
    chan_width = ps[list(ps.keys())[0]].frequency.values
    pol_coords = ps[list(ps.keys())[0]].polarization
    time_coords = np.mean(ps[list(ps.keys())[0]].time.values)

    # print(freq_coords.shape,input_parms['data_sel']['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_1'])

    # print(img_xds)
    img_xds = xr.Dataset()
    img_xds = make_empty_sky_image(
        img_xds,
        grid_parms["phase_direction"]["data"],
        grid_parms["image_size"],
        grid_parms["cell_size"],
        freq_coords.values,
        chan_width,
        pol_coords,
        time_coords,
        data_group_name="mosaic",
    )

    # for ms_xds in [list(ps.values())[0]]:
    for ms_xds in ps.values():
        # print(ms_xds)
        data_group_out = _phase_shift_vis_ds(
            ms_xds, shift_parms=shift_parms, sel_parms={}
        )

        # data_group_out = _make_imaging_weights(ms_xds,grid_parms=grid_parms,imaging_weights_parms={'weighting':'briggs','robust':0.6},sel_parms={"data_group_in":data_group_out})
        data_group_out = _make_imaging_weights(
            ms_xds,
            grid_parms=grid_parms,
            imaging_weights_parms={"weighting": "natural", "robust": 0.6},
            sel_parms={"data_group_in": data_group_out},
        )

        gcf_parms = {}
        gcf_parms["function"] = "casa_airy"
        gcf_parms["list_dish_diameters"] = np.array([10.7])
        gcf_parms["list_blockage_diameters"] = np.array([0.75])

        # print(ms_xds.attrs['antenna_xds'])
        unique_ant_indx = ms_xds.attrs["antenna_xds"].DISH_DIAMETER.values
        unique_ant_indx[unique_ant_indx == 12.0] = 0

        gcf_parms["unique_ant_indx"] = unique_ant_indx.astype(int)
        gcf_parms["phase_direction"] = grid_parms["phase_direction"]
        gcf_xds = _make_gridding_convolution_function(
            ms_xds, gcf_parms, grid_parms, sel_parms={"data_group_in": data_group_out}
        )
        # print(gcf_xds)
        # print(ms_xds['WEIGHT_IMAGING'])

        _make_aperture_grid(
            ms_xds,
            gcf_xds,
            img_xds,
            vis_sel_parms={"data_group_in": data_group_out},
            img_sel_parms={"data_group_in": "mosaic"},
            grid_parms=grid_parms,
        )

        _make_uv_sampling_grid(
            ms_xds,
            gcf_xds,
            img_xds,
            vis_sel_parms={"data_group_in": data_group_out},
            img_sel_parms={"data_group_in": "mosaic"},
            grid_parms=grid_parms,
        )

        _make_visibility_grid(
            ms_xds,
            gcf_xds,
            img_xds,
            vis_sel_parms={"data_group_in": data_group_out},
            img_sel_parms={"data_group_in": "mosaic"},
            grid_parms=grid_parms,
        )

    _fft_norm_img_xds(
        img_xds,
        gcf_xds,
        grid_parms,
        norm_parms={},
        sel_parms={"data_group_in": "mosaic", "data_group_out": "mosaic"},
    )

    # Tranform uv-space -> lm-space (sky)

    return img_xds


"""

    ps_name = '/Users/jsteeb/Dropbox/Data/Antennae_North.cal.lsrk.vis.zarr'
    sel_parms = {}
    sel_parms['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_1'] = {}
#    sel_parms['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_2'] = {}
#    sel_parms['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_3'] = {}
#    sel_parms['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_13'] = {}
#    sel_parms['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_4'] = {}
#    sel_parms['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_5'] = {}
#    sel_parms['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_6'] = {}
#    sel_parms['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_7'] = {}
#    sel_parms['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_8'] = {}
#    sel_parms['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_9'] = {}
#    sel_parms['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_10'] = {}
#    sel_parms['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_11'] = {}
    sel_parms['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_12'] = {}

#    sel_parms['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_14'] = {}
#    sel_parms['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_15'] = {}
#    sel_parms['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_16'] = {}
#    sel_parms['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_17'] = {}
#    sel_parms['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_18'] = {}
#    sel_parms['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_19'] = {}
#    sel_parms['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_20'] = {}
#    sel_parms['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_21'] = {}
#    sel_parms['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_22'] = {}
#    sel_parms['Antennae_North.cal.lsrk_ddi_0_intent_OBSERVE_TARGET#ON_SOURCE_field_id_23'] = {}

    ps = load_processing_set(ps_name,sel_parms)

"""
